package wacc

import scala.collection.mutable.ListBuffer
import cats.data.Chain
import cats.syntax.foldable._

object asmGenerator {
  import microWacc._
  import assemblyIR._
  import assemblyIR.Size._
  import assemblyIR.RegName._
  import types._
  import sizeExtensions._
  import lexer.escapedChars

  abstract case class Error() {
    def strLabel: String
    def errStr: String
    def errLabel: String

    def stringDef: Chain[AsmLine] = Chain(
      Directive.Int(errStr.size),
      LabelDef(strLabel),
      Directive.Asciz(errStr)
    )
  }
  object zeroDivError extends Error {
    // TODO: is this bad? Can we make an error case class/some other structure?
    def strLabel = ".L._errDivZero_str0"
    def errStr = "fatal error: division or modulo by zero"
    def errLabel = ".L._errDivZero"
  }

  private val RAX = Register(Q64, AX)
  private val EAX = Register(D32, AX)
  private val RDI = Register(Q64, DI)
  private val RIP = Register(Q64, IP)
  private val RBP = Register(Q64, BP)
  private val RSI = Register(Q64, SI)
  private val RDX = Register(Q64, DX)
  private val RCX = Register(Q64, CX)
  private val argRegs = List(DI, SI, DX, CX, R8, R9)

  extension [T](chain: Chain[T])
    def +(item: T): Chain[T] = chain.append(item)

    def concatAll(chains: Chain[T]*): Chain[T] =
      chains.foldLeft(chain)(_ ++ _)

  class LabelGenerator {
    var labelVal = -1
    def getLabel(): String = {
      labelVal += 1
      s".L$labelVal"
    }
    def getLabel(target: CallTarget): String = target match {
      case Ident(v, _)   => s"wacc_$v"
      case Builtin(name) => s"_$name"
    }
  }

  def generateAsm(microProg: Program): Chain[AsmLine] = {
    given stack: Stack = Stack()
    given strings: ListBuffer[String] = ListBuffer[String]()
    given labelGenerator: LabelGenerator = LabelGenerator()
    val Program(funcs, main) = microProg

    val progAsm = Chain(LabelDef("main")).concatAll(
      funcPrologue(),
      main.foldMap(generateStmt(_)),
      Chain.one(Xor(RAX, RAX)),
      funcEpilogue(),
      generateBuiltInFuncs(),
      funcs.foldMap(generateUserFunc(_))
    )

    val strDirs = strings.toList.zipWithIndex.foldMap { case (str, i) =>
      Chain(
        Directive.Int(str.size),
        LabelDef(s".L.str$i"),
        Directive.Asciz(str.escaped)
      )
    } ++ zeroDivError.stringDef

    Chain(
      Directive.IntelSyntax,
      Directive.Global("main"),
      Directive.RoData
    ).concatAll(
      strDirs,
      Chain.one(Directive.Text),
      progAsm
    )
  }

  private def wrapBuiltinFunc(labelName: String, funcBody: Chain[AsmLine]): Chain[AsmLine] = {
    var chain = Chain.one[AsmLine](LabelDef(labelName))
    chain ++= funcPrologue()
    chain ++= funcBody
    chain ++= funcEpilogue()
    chain
  }

  private def generateUserFunc(func: FuncDecl)(using
      strings: ListBuffer[String],
      labelGenerator: LabelGenerator
  ): Chain[AsmLine] = {
    given stack: Stack = Stack()
    // Setup the stack with param 7 and up
    func.params.drop(argRegs.size).foreach(stack.reserve(_))
    var chain = Chain.one[AsmLine](LabelDef(labelGenerator.getLabel(func.name)))
    chain ++= funcPrologue()
    // Push the rest of params onto the stack for simplicity
    argRegs.zip(func.params).foreach { (reg, param) =>
      chain += stack.push(param, Register(Q64, reg))
    }
    chain ++= func.body.foldMap(generateStmt(_))
    // No need for epilogue here since all user functions must return explicitly
    chain
  }

  private def generateBuiltInFuncs()(using
      labelGenerator: LabelGenerator
  ): Chain[AsmLine] = {
    var chain = Chain.empty[AsmLine]

    chain ++= wrapBuiltinFunc(
      labelGenerator.getLabel(Builtin.Exit),
      Chain(stackAlign, assemblyIR.Call(CLibFunc.Exit))
    )

    chain ++= wrapBuiltinFunc(
      labelGenerator.getLabel(Builtin.Printf),
      Chain(
        stackAlign,
        assemblyIR.Call(CLibFunc.PrintF),
        Xor(RDI, RDI),
        assemblyIR.Call(CLibFunc.Fflush)
      )
    )

    chain ++= wrapBuiltinFunc(
      labelGenerator.getLabel(Builtin.PrintCharArray),
      Chain(
        stackAlign,
        Load(RDX, IndexAddress(RSI, KnownType.Int.size.toInt)),
        Move(Register(D32, SI), MemLocation(RSI, D32)),
        assemblyIR.Call(CLibFunc.PrintF),
        Xor(RDI, RDI),
        assemblyIR.Call(CLibFunc.Fflush)
      )
    )

    chain ++= wrapBuiltinFunc(
      labelGenerator.getLabel(Builtin.Malloc),
      Chain(stackAlign, assemblyIR.Call(CLibFunc.Malloc))
      // Out of memory check is optional
    )

    chain ++= wrapBuiltinFunc(labelGenerator.getLabel(Builtin.Free), Chain.empty)

    chain ++= wrapBuiltinFunc(
      labelGenerator.getLabel(Builtin.Read),
      Chain(
        stackAlign,
        Subtract(Register(Q64, SP), ImmediateVal(8)),
        Push(RSI),
        Load(RSI, MemLocation(Register(Q64, SP), Q64)),
        assemblyIR.Call(CLibFunc.Scanf),
        Pop(RAX)
      )
    )

    chain ++= Chain(
      // TODO can this be done with a call to generateStmt?
      // Consider other error cases -> look to generalise
      LabelDef(zeroDivError.errLabel),
      stackAlign,
      Load(RDI, IndexAddress(RIP, LabelArg(zeroDivError.strLabel))),
      assemblyIR.Call(CLibFunc.PrintF),
      Move(RDI, ImmediateVal(-1)),
      assemblyIR.Call(CLibFunc.Exit)
    )

    chain
  }

  private def generateStmt(stmt: Stmt)(using
      stack: Stack,
      strings: ListBuffer[String],
      labelGenerator: LabelGenerator
  ): Chain[AsmLine] = {
    var chain = Chain.empty[AsmLine]
    chain += Comment(stmt.toString)
    stmt match {
      case Assign(lhs, rhs) =>
        lhs match {
          case ident: Ident =>
            if (!stack.contains(ident)) chain += stack.reserve(ident)
            chain ++= evalExprOntoStack(rhs)
            chain += stack.pop(RAX)
            chain += Move(stack.accessVar(ident), RAX)
          case ArrayElem(x, i) =>
            chain ++= evalExprOntoStack(rhs)
            chain ++= evalExprOntoStack(i)
            chain ++= evalExprOntoStack(x)
            chain += stack.pop(RAX)
            chain += stack.pop(RCX)
            chain += stack.pop(RDX)

            chain += Move(
              IndexAddress(RAX, KnownType.Int.size.toInt, RCX, x.ty.elemSize.toInt),
              Register(x.ty.elemSize, DX)
            )
        }

      case If(cond, thenBranch, elseBranch) =>
        val elseLabel = labelGenerator.getLabel()
        val endLabel = labelGenerator.getLabel()

        chain ++= evalExprOntoStack(cond)
        chain += stack.pop(RAX)
        chain += Compare(RAX, ImmediateVal(0))
        chain += Jump(LabelArg(elseLabel), Cond.Equal)

        chain ++= stack.withScope(() => thenBranch.foldMap(generateStmt))
        chain += Jump(LabelArg(endLabel))
        chain += LabelDef(elseLabel)

        chain ++= stack.withScope(() => elseBranch.foldMap(generateStmt))
        chain += LabelDef(endLabel)

      case While(cond, body) =>
        val startLabel = labelGenerator.getLabel()
        val endLabel = labelGenerator.getLabel()

        chain += LabelDef(startLabel)
        chain ++= evalExprOntoStack(cond)
        chain += stack.pop(RAX)
        chain += Compare(RAX, ImmediateVal(0))
        chain += Jump(LabelArg(endLabel), Cond.Equal)

        chain ++= stack.withScope(() => body.foldMap(generateStmt))
        chain += Jump(LabelArg(startLabel))
        chain += LabelDef(endLabel)

      case call: microWacc.Call =>
        chain ++= generateCall(call, isTail = false)

      case microWacc.Return(expr) =>
        expr match {
          case call: microWacc.Call =>
            chain ++= generateCall(call, isTail = true) // tco
          case _ =>
            chain ++= evalExprOntoStack(expr)
            chain += stack.pop(RAX)
            chain ++= funcEpilogue()
        }
    }

    chain
  }

  private def evalExprOntoStack(expr: Expr)(using
      stack: Stack,
      strings: ListBuffer[String],
      labelGenerator: LabelGenerator
  ): Chain[AsmLine] = {
    var chain = Chain.empty[AsmLine]
    val stackSizeStart = stack.size
    expr match {
      case IntLiter(v)  => chain += stack.push(KnownType.Int.size, ImmediateVal(v))
      case CharLiter(v) => chain += stack.push(KnownType.Char.size, ImmediateVal(v.toInt))
      case ident: Ident => chain += stack.push(ident.ty.size, stack.accessVar(ident))

      case array @ ArrayLiter(elems) =>
        expr.ty match {
          case KnownType.String =>
            strings += elems.collect { case CharLiter(v) => v }.mkString
            chain += Load(RAX, IndexAddress(RIP, LabelArg(s".L.str${strings.size - 1}")))
            chain += stack.push(Q64, RAX)
          case ty =>
            chain ++= generateCall(
              microWacc.Call(Builtin.Malloc, List(IntLiter(array.heapSize))),
              isTail = false
            )
            chain += stack.push(Q64, RAX)
            // Store the length of the array at the start
            chain += Move(MemLocation(RAX, D32), ImmediateVal(elems.size))
            elems.zipWithIndex.foldMap { (elem, i) =>
              chain ++= evalExprOntoStack(elem)
              chain += stack.pop(RCX)
              chain += stack.pop(RAX)
              chain += Move(IndexAddress(RAX, 4 + i * ty.elemSize.toInt), Register(ty.elemSize, CX))
              chain += stack.push(Q64, RAX)
            }
        }

      case BoolLiter(true) =>
        chain += stack.push(KnownType.Bool.size, ImmediateVal(1))
      case BoolLiter(false) =>
        chain += Xor(RAX, RAX)
        chain += stack.push(KnownType.Bool.size, RAX)
      case NullLiter() =>
        chain += stack.push(KnownType.Pair(?, ?).size, ImmediateVal(0))
      case ArrayElem(x, i) =>
        chain ++= evalExprOntoStack(x)
        chain ++= evalExprOntoStack(i)
        chain += stack.pop(RCX)
        chain += stack.pop(RAX)
        // + Int because we store the length of the array at the start
        chain += Move(
          Register(x.ty.elemSize, AX),
          IndexAddress(RAX, KnownType.Int.size.toInt, RCX, x.ty.elemSize.toInt)
        )
        chain += stack.push(x.ty.elemSize, RAX)
      case UnaryOp(x, op) =>
        chain ++= evalExprOntoStack(x)
        op match {
          case UnaryOperator.Chr | UnaryOperator.Ord => // No op needed
          case UnaryOperator.Len =>
            chain += stack.pop(RAX)
            chain += Move(EAX, MemLocation(RAX, D32))
            chain += stack.push(D32, RAX)
          case UnaryOperator.Negate =>
            chain += Negate(stack.head)
          case UnaryOperator.Not =>
            chain += Xor(stack.head, ImmediateVal(1))
        }

      case BinaryOp(x, y, op) =>
        val destX = Register(x.ty.size, AX)
        chain ++= evalExprOntoStack(y)
        chain ++= evalExprOntoStack(x)
        chain += stack.pop(RAX)

        op match {
          case BinaryOperator.Add =>
            chain += Add(stack.head, destX)
          case BinaryOperator.Sub =>
            chain += Subtract(destX, stack.head)
            chain += stack.drop()
            chain += stack.push(destX.size, RAX)
          case BinaryOperator.Mul =>
            chain += Multiply(destX, stack.head)
            chain += stack.drop()
            chain += stack.push(destX.size, RAX)

          case BinaryOperator.Div =>
            chain += Compare(stack.head, ImmediateVal(0))
            chain += Jump(LabelArg(zeroDivError.errLabel), Cond.Equal)
            chain += CDQ()
            chain += Divide(stack.head)
            chain += stack.drop()
            chain += stack.push(destX.size, RAX)

          case BinaryOperator.Mod =>
            chain += CDQ()
            chain += Divide(stack.head)
            chain += stack.drop()
            chain += stack.push(destX.size, RDX)

          case BinaryOperator.Eq        => chain ++= generateComparison(destX, Cond.Equal)
          case BinaryOperator.Neq       => chain ++= generateComparison(destX, Cond.NotEqual)
          case BinaryOperator.Greater   => chain ++= generateComparison(destX, Cond.Greater)
          case BinaryOperator.GreaterEq => chain ++= generateComparison(destX, Cond.GreaterEqual)
          case BinaryOperator.Less      => chain ++= generateComparison(destX, Cond.Less)
          case BinaryOperator.LessEq    => chain ++= generateComparison(destX, Cond.LessEqual)
          case BinaryOperator.And       => chain += And(stack.head, destX)
          case BinaryOperator.Or        => chain += Or(stack.head, destX)
        }

      case call: microWacc.Call =>
        chain ++= generateCall(call, isTail = false)
        chain += stack.push(call.ty.size, RAX)
    }

    assert(stack.size == stackSizeStart + 1)
    chain ++= zeroRest(MemLocation(stack.head.pointer, Q64), expr.ty.size)
    chain
  }

  private def generateCall(call: microWacc.Call, isTail: Boolean)(using
      stack: Stack,
      strings: ListBuffer[String],
      labelGenerator: LabelGenerator
  ): Chain[AsmLine] = {
    var chain = Chain.empty[AsmLine]
    val microWacc.Call(target, args) = call

    argRegs.zip(args).foldMap { (reg, expr) =>
      chain ++= evalExprOntoStack(expr)
      chain += stack.pop(Register(Q64, reg))
    }

    args.drop(argRegs.size).foldMap {
      chain ++= evalExprOntoStack(_)
    }

    // Tail Call Optimisation (TCO)
    if (isTail) {
      chain += Jump(LabelArg(labelGenerator.getLabel(target))) // tail call
    } else {
      chain += assemblyIR.Call(LabelArg(labelGenerator.getLabel(target))) // regular call
    }

    if (args.size > argRegs.size) {
      chain += stack.drop(args.size - argRegs.size)
    }

    chain
  }

  private def generateComparison(destX: Register, cond: Cond)(using
      stack: Stack
  ): Chain[AsmLine] = {
    var chain = Chain.empty[AsmLine]

    chain += Compare(destX, stack.head)
    chain += Set(Register(B8, AX), cond)
    chain ++= zeroRest(RAX, B8)
    chain += stack.drop()
    chain += stack.push(B8, RAX)

    chain
  }

  private def funcPrologue(): Chain[AsmLine] = {
    var chain = Chain.empty[AsmLine]
    chain += Push(RBP)
    chain += Move(RBP, Register(Q64, SP))
    chain
  }

  private def funcEpilogue(): Chain[AsmLine] = {
    var chain = Chain.empty[AsmLine]
    chain += Move(Register(Q64, SP), RBP)
    chain += Pop(RBP)
    chain += assemblyIR.Return()
    chain
  }

  private def stackAlign: AsmLine = And(Register(Q64, SP), ImmediateVal(-16))
  private def zeroRest(dest: Dest, size: Size): Chain[AsmLine] = size match {
    case Q64 | D32 => Chain.empty
    case _         => Chain.one(And(dest, ImmediateVal((1 << (size.toInt * 8)) - 1)))
  }

  private val escapedCharsMapping = escapedChars.map { case (k, v) => v -> s"\\$k" }
  extension (s: String) {
    private def escaped: String =
      s.flatMap(c => escapedCharsMapping.getOrElse(c, c.toString))
  }
}
